#include "Card.h"
#include "Solver.h"
#include "Monster.h"
#include "Main.h"

#include "Chars.h"
#include "Vram.h"
#include "Sprite.h"
#include "Status.h"
#include "Sound.h"
#include "VVram.h"

extern void WaitTimer(byte t);

Card[CardCount] Cards;
ptr<Card> pHeldCard;
ptr<Card> pThrownCard;
byte RemainCardCount;

byte CardX, CardY;
sbyte CardDx, CardDy;

static byte FromPosition(byte a)
{
    return a + a + a + 1;
}

void InitCards(ptr<CardPosition> pPosition, byte count) 
{
    ptr<Card> pCard;
    byte i;

    RemainCardCount = count;
    pCard = Cards;
    i = 0;
    while (count != 0 && i < CardCount) {
        pCard->status = Card_Status_Floor | pPosition->number;
        pCard->x = FromPosition(pPosition->position >> 4);
        pCard->y = FromPosition(pPosition->position & 0x0f);
        pCard->old = 0;
        // VPut2(pCard->x, pCard->y, Pattern_Card);
        ++pCard;
        ++pPosition;
        --count;
        ++i;
    }
    while (i < CardCount) {
        pCard->status = Card_Status_None;
        ++pCard;
        ++i;
    }

    pHeldCard = nullptr;
    pThrownCard = nullptr;
}

void PickCard() 
{
    byte x, y;
    ptr<Card> pCard;

    if (pHeldCard == nullptr) {
        x = (SolverX + 4) >> 3;
        y = (SolverY + 4) >> 3;

        for (pCard: Cards) {
            if (
                (pCard->status & Card_Status_Mask) == Card_Status_Floor &&
                pCard->x == x && pCard->y == y
            ) {
                pHeldCard = pCard;
                pCard->status = (pCard->status & Card_Number_Mask) | Card_Status_Held;
                Sound_Get();
                DrawAll();
                PrintHeldCard();
                return;
            }
        }
    }
}

static bool TestToMove(byte x, byte y) 
{
    if (TestMap2(x, y)) {
        byte gridX, gridY;
        ptr<Card> pCard;
        
        gridX = x >> 3;
        gridY = y >> 3;

        for (pCard: Cards) {
            if ((pCard->status & Card_Status_Mask) == Card_Status_Floor) {
                byte xDiff, yDiff;
                xDiff = gridX - pCard->x;
                yDiff = gridY - pCard->y;
                if (xDiff + 1 < 3 && yDiff + 1 < 3) {
                    return false;
                }
            }
        }
        return true;
    }
    return false;
}

void ThrowCard(byte x, byte y, sbyte dx, sbyte dy) 
{
    x = (x + 3) & 0xf8;
    y = (y + 3) & 0xf8;
    if (pHeldCard != nullptr && pThrownCard == nullptr && TestToMove(x, y)) {
        CardX = x;
        CardY = y;
        CardDx = dx << 3;
        CardDy = dy << 3;
        pThrownCard = pHeldCard;
        pHeldCard = nullptr;
        Sound_Get();
        pThrownCard->status = (pThrownCard->status & Card_Number_Mask) | Card_Status_Move;
        // ShowSprite(Sprite, CardX, CardY, Char_Card);
        // UpdateSprites();
        PrintHeldCard();
    }
}

void TestMatching(ptr<Card> p) 
{
    ptr<Card> pCard;
      
    for (pCard: Cards) {
        if (pCard == p) continue;
        if ((pCard->status & Card_Status_Mask) != Card_Status_Floor) continue;

        if (
            (p->x == pCard->x && p->y - pCard->y + 2 < 5) ||
            (p->y == pCard->y && p->x - pCard->x + 2 < 5)
        ) {
            byte number, otherNumber;
            
            number =  p->status & Card_Number_Mask;
            p->status = Card_Status_Open | number;
            // VPut2(p->x, p->y, Char_Card + number);
            otherNumber = pCard->status & Card_Number_Mask;
            pCard->status = Card_Status_Open | otherNumber;
            // VPut2(pCard->x, pCard->y, Char_Card + otherNumber);
            DrawAll();
            if (number == otherNumber) {
                Sound_Hit();
                p->status = Card_Status_None;
                // VErase2(p->x, p->y);
                pCard->status = Card_Status_None;
                // VErase2(pCard->x, pCard->y);
                RemainCardCount -= 2;
                AddScore(10);   // 100pts
                WaitTimer(15);
                return;
            }
            else {
                Sound_Miss();
                p->status = Card_Status_Floor | number;
                pCard->status = Card_Status_Floor | otherNumber;
                // VPut2(p->x, p->y, Char_Card);
                // VPut2(pCard->x, pCard->y, Char_Card);
                WaitTimer(15);
                DrawAll();
            }
        }
    }
}


void MoveCard()
{
    if (pThrownCard != nullptr) {
        byte newX, newY;
        newX = CardX + CardDx;
        newY = CardY + CardDy;
        if (TestToMove(newX, newY) && !Monster_TestHit(newX, newY)) {
            CardX = newX;
            CardY = newY;
            pThrownCard->x = (CardX + 4) >> 3;
            pThrownCard->y = (CardY + 4) >> 3;
            // ShowSprite(Sprite, newX, newY, Pattern_Card);
        }
        else {
            pThrownCard->status = (pThrownCard->status & Card_Number_Mask) | Card_Status_Floor;
            pThrownCard->x = (CardX + 4) >> 3;
            pThrownCard->y = (CardY + 4) >> 3;
            TestMatching(pThrownCard);
            pThrownCard = nullptr;
            // VPut2(pThrownCard->x, pThrownCard->y, Pattern_Card);
            // DrawAll();
            // HideSprite(Sprite);
        }
    }
}


void DrawCards()
{
    ptr<Card> pCard;
    byte status;
    for (pCard: Cards) {
        status = pCard->status & Card_Status_Mask;
        if (status == Card_Status_None || status == Card_Status_Held) {
        }
        else {
            byte c, x, y;
            ptr<byte> pVVram;
            bool changed;
            c = Char_Card;
            x = pCard->x;
            y = pCard->y;
            if (status == Card_Status_Move) {
                changed = true;
            }
            else if (status != Card_Status_Floor) {
                c += (pCard->status & Card_Number_Mask) << 2;
                changed = pCard->status != pCard->old;
            }
            pVVram = VVramPtr(x, y);
            repeat (2) {
                repeat (2) {
                    *pVVram = c;
                    ++pVVram;
                    ++c;
                }
                pVVram += VVramWidth - 2;
            }
            pCard->old = pCard->status;
            if (changed) {
                SetRowFlags(y, 3);
            }
        }
    }
}