#include "Status.h"
#include "Vram.h"
#include "Stage.h"
#include "Main.h"
#include "Sprite.h"
#include "Card.h"
#include "Chars.h"
#include "Print.h"

constexpr byte Status_X = 32;
constexpr byte Char_Remain = Char_Solver;

static ptr<byte> PrintS(ptr<byte> pVram, ptr<byte> p)
{
    byte c;
    while ((c = *p) != 0) {
        pVram = PrintC(pVram, c);
        ++p;
    }
}

void PrintStatus() 
{
    PrintS(Vram + VramRowSize + Status_X, "SCORE");
    PrintS(Vram + VramRowSize * 4 + Status_X, "HIGH");
    PrintS(Vram + VramRowSize * 5 + Status_X, "SCORE");
    PrintS(Vram + VramRowSize * 8 + Status_X, "STAGE");
    PrintByteNumber2(Vram + VramRowSize * 9 + Status_X + 3, CurrentStage + 1);
    PrintS(Vram + VramRowSize * 11 + Status_X, "TIME");

    {
        ptr<byte> pVram;
        byte count;
        pVram = Vram + VramRowSize * 21 + Status_X;

        for (count = 1; count < RemainSolverCount; ++count) {
            Print2x2(pVram, Char_Remain);
            pVram += 2;
        }
        for (; count < 3; ++count) {
            Erase2x2(pVram);
            pVram += 2;
        }
    }
    PrintScore();
    PrintTime();
    PrintHeldCard();
}

void PrintScore()
{
    ptr<byte> pVram;
    
    pVram = PrintNumber5(Vram + VramRowSize * 2 + Status_X, Score);
    PrintC(pVram, '0');

    pVram = PrintNumber5(Vram + VramRowSize * 6 + Status_X, HiScore);
    PrintC(pVram, '0');
}

void PrintTime()
{
    PrintNumber3(Vram + VramRowSize * 12 + Status_X + 1, StageTime);
}

void PrintHeldCard() 
{
    if (pHeldCard != nullptr) {
        byte pattern;
        
        pattern = Char_Card + ((pHeldCard->status & Card_Number_Mask) << 2);
        Print2x2(Vram + VramRowSize * 18 + Status_X, pattern);
    }
    else {
        Erase2x2(Vram + VramRowSize * 18 + Status_X);
    }
}

void PrintGameOver()
{
    PrintS(Vram + VramRowSize * 12 + 11, "GAME OVER");
}

void PrintTimeUp() 
{
    PrintS(Vram + VramRowSize * 12 + 12, "TIME UP");
}

void Title()
{
    ClearScreen();
    // {
    //     ptr<byte> pByte;
    //     word count;

    //     pByte = Backgnd;
    //     count = StageWidth * StageHeight;
    //     do {
    //         *pByte = Char_Space;
    //         ++pByte;
    //         --count;
    //     } while (count !=0);
    // }
    {
        byte c;
        ptr<byte> pVram;
        
        c = Char_Title;
        pVram = Vram + VramRowSize * 1 + 4;
        repeat(6) {
            repeat (4) {
                repeat (4) {
                    VPut(pVram, c);
                    ++pVram;
                    ++c;
                }
                pVram += VramRowSize - 4;
            }
            pVram += 4 - VramRowSize * 4;
        }
    }
    Print2x2(Vram + VramRowSize * 8 + 9, Char_Solver);
    Print2x2(Vram + VramRowSize * 11 + 9, Char_Monster);
    Print2x2(Vram + VramRowSize * 14 + 9, Char_Card);

    PrintStatus();
    PrintS(Vram + VramRowSize * 6 + 7, "--- CHARACTER ---");
    PrintS(Vram + VramRowSize * 9 + 13, "SOLVER");
    PrintS(Vram + VramRowSize * 12 + 13, "MONSTER");
    PrintS(Vram + VramRowSize * 15 + 13, "CARD");
    PrintS(Vram + VramRowSize * 21 + 9, "PUSH SPACE KEY");
    PrintS(Vram + VramRowSize * 22 + 9, "OR SHIFT KEY");
}